# Copyright (c) 2022 PicsArt, Inc.
# All rights reserved. Use of this source code is governed by a
# MIT-style license that can be found in the LICENSE file.

"""
Helper module for processing yaml files
"""
import copy
import os

from collections.abc import MutableMapping
from functools import lru_cache

import yaml

from cppbind import Error
from cppbind.common.config import config, PROJECT_CONFIG_DIR
from cppbind.utils import list_files, filter_by_glob


class YamlNode(MutableMapping):
    """
    Class which implements MutableMapping interface to act like a dict
    while keeping additional information. This is implemented to be able to keep
    each yaml node line and file information while loading yaml files.
    """
    def __init__(self, value, line_num=None, file=None):
        self.value = value
        self.line_number = line_num
        self.file = file

    def __getitem__(self, key):
        return self.value[key]

    def __setitem__(self, key, val):
        self.value[key] = val

    def __delitem__(self, key):
        del self.value[key]

    def __iter__(self):
        return iter(self.value)

    def __len__(self):
        return len(self.value)

    def __str__(self):
        return str(self.value)

    def __contains__(self, item):
        return item in self.value

    def is_of_type(self, cls):
        """
        Method to check current node real value type.
        """
        return isinstance(self.value, cls)

    def __bool__(self):
        return bool(self.value)

    def __eq__(self, other):
        """
        This method is overloaded, to avoid usage of 'value' property
        when comparing YamlNode object with other simple object
        """
        if isinstance(other, YamlNode):
            return self.value == other.value
        return self.value == other


class YamlKeyDuplicationError(Exception):
    """
    Exception subclass to raise duplication errors when a key is redefined in yaml file.
    """


@lru_cache
def init_yaml_list(dirs):
    yaml_files = []
    for d in dirs:
        yaml_files.extend(list_files(d, ('.yaml', '.yml')))
    return yaml_files


class MyLoader(yaml.SafeLoader):
    """YAML MyLoader with `!include` constructor."""

    # the list of directories where included yaml files need to be searched
    custom_dirs = [PROJECT_CONFIG_DIR]
    if hasattr(config.application, 'custom_config_dir'):
        custom_dirs.append(config.application.custom_config_dir)

    def __init__(self, stream):
        """Initialise MyLoader."""

        try:
            self._root = os.path.split(stream.name)[0]
        except AttributeError:
            self._root = os.path.curdir

        self.dirs = [self._root]
        self.yaml_files = init_yaml_list(dirs=tuple(self.custom_dirs + [os.path.abspath(os.path.curdir)]))
        super().__init__(stream)

    def construct_yaml_map(self, node):
        """
        Original class is responsible for constructing yaml loader result nodes.
        In this overridden version dict is being replaced with custom class object.
        """
        yield self.construct_mapping(node)

    def construct_mapping(self, node, deep=False):
        """
        Original method is overridden to be able to get a custom class object
        instead of dict when loading yaml.
        """
        mapping = {}
        for key_node, value_node in node.value:
            key = self.construct_object(key_node, deep=deep)
            value = self.construct_object(value_node, deep=deep)
            mapping[key] = YamlNode(value,
                                    key_node.start_mark.line + 1,
                                    key_node.start_mark.name)
        return mapping


class UniqueKeyLoader(yaml.SafeLoader):
    """
    Custom yaml loader to raise an error when duplicate key is found in yaml file.
    """
    def construct_mapping(self, node, deep=False):
        mapping = []
        for key_node, _ in node.value:
            key = self.construct_object(key_node, deep=deep)
            if key in mapping:
                raise YamlKeyDuplicationError(f"Yaml error: duplicate key: {key}")
            mapping.append(key)
        return super().construct_mapping(node, deep)


def join_nodes(rdata, edata):
    """Join nodes defined after the !join constructor"""
    if rdata is None:
        rdata = copy.copy(edata)
    elif has_type(rdata, dict):
        rdata.update(edata)
    else:
        rdata += edata
    return to_value(rdata)


def construct_include(loader, node):
    """Include file referenced at node."""
    try:
        entries = loader.construct_sequence(node)
    except Exception:
        entries = [loader.construct_scalar(node)]

    search_dirs = loader.dirs + loader.custom_dirs

    def load_entry(entry):
        filename = entry
        sub_index = entry.rfind('&')
        sub_node = None
        if sub_index != -1:
            sub_node = entry[sub_index + 1::]
            sub_node = sub_node.split('.')
            filename = entry[:sub_index]

        if os.path.isabs(filename):
            filenames = [filename]
        else:
            filenames = [os.path.abspath(os.path.join(f"{path}/**/", filename))
                         for path in search_dirs]
        extension = os.path.splitext(filename)[1].lstrip('.')

        rdata = None

        file_paths = set(item for pattern in filenames for item in filter_by_glob(loader.yaml_files, pattern))
        for file_path in file_paths:
            sub_yaml = None
            try:
                if extension in ('yaml', 'yml'):
                    sub_yaml = load_yaml(file_path)
                    if sub_node is not None:
                        for nselect in sub_node:
                            if isinstance(sub_yaml, list):
                                nselect = int(nselect)
                            sub_yaml = sub_yaml[nselect]
                else:
                    Error.critical(f'Can only include yaml file: {file_path} file has not supported extension "{extension}"')
            except OSError as err:
                Error.critical(f"Cannot read file {file_path}: {err}")

            rdata = join_nodes(rdata, sub_yaml)

        return rdata

    rdata = None
    for entry in entries:
        edata = load_entry(entry)
        rdata = join_nodes(rdata, edata)
    return rdata


def construct_join(loader, node):
    """join nodes."""

    original = loader.construct_object

    def construct_object(node, deep=False):
        if isinstance(node, yaml.MappingNode):
            return loader.construct_mapping(node, deep)
        if isinstance(node, yaml.SequenceNode):
            return loader.construct_sequence(node, deep)
        return original(node, deep)

    loader.construct_object = construct_object
    try:
        entries = loader.construct_sequence(node)
        rdata = None
        for entry in entries:
            rdata = join_nodes(rdata, entry)

        return rdata
    finally:
        loader.construct_object = original


def construct_concat(loader, node):
    """Custom constructor to create a node by concatenating string values of input nodes"""

    if not isinstance(node, yaml.SequenceNode):
        Error.critical(f"`concat` yaml constructor can operate only on sequence node")

    entries = loader.construct_sequence(node)
    res = ''
    for entry in entries:
        if not isinstance(entry, str):
            Error.critical(f"Items of `concat` yaml constructor must be strings, {entry} has type {type(entry)}")
        res = res + entry

    return res


yaml.add_constructor('!include', construct_include, MyLoader)
yaml.add_constructor('!join', construct_join, MyLoader)
yaml.add_constructor('!concat', construct_concat, MyLoader)

# add additional constructor to force MyLoader loader
# to call our custom construct_yaml_map method when constructing nodes.
MyLoader.add_constructor('tag:yaml.org,2002:map', MyLoader.construct_yaml_map)


@lru_cache
def load_yaml(file_path):
    """Load yaml file in specific dirs using MyLoader custom loader"""
    try:
        with open(file_path) as file:
            return yaml.load(file, MyLoader)
    except OSError as err:
        Error.critical(f"Cannot read file {file_path}: {err}")


def has_type(obj, type_cls):
    """
    A function to determine whether an object is an instance of given type,
    or is an instance of YamlNode which has value of given type.
    Used to eliminate some long checks.
    """
    return obj.is_of_type(type_cls) if isinstance(obj, YamlNode) else isinstance(obj, type_cls)


def to_value(obj):
    """
    A function to return real value of object
    """
    if isinstance(obj, YamlNode):
        return obj.value
    return obj


def yaml_info_struct_to_dict(struct):
    """
    A function to rebuild dict from nested YamlNode object.
    """
    if has_type(struct, dict):
        return {k: yaml_info_struct_to_dict(v) for k, v in struct.items()}
    if has_type(struct, list):
        return [yaml_info_struct_to_dict(i) for i in struct]
    if isinstance(struct, YamlNode):
        return struct.value
    return struct



# Copyright (c) 2022 PicsArt, Inc.
# All rights reserved. Use of this source code is governed by a
# MIT-style license that can be found in the LICENSE file.
import types
from functools import lru_cache

from cached_property import cached_property

from cppbind.cxx_exposed import CXXExposedType, CXXRunnerExposedType
from cppbind.ir.exec_rules import RunRule
from cppbind.utils.clang import replace_template_choice


@lru_cache(maxsize=512)
def create_type_info(runner: RunRule, cxx_exposed_type: CXXExposedType):
    return TypeInfo(runner, cxx_exposed_type)


class TypeInfo:

    def __init__(self, runner, cxx_exposed_type):
        self._runner = runner
        self._cxx_exposed_type = cxx_exposed_type
        # get canonical_pointee_type to be able to find it's context(cxx type might be a typedef, pointer etc.)
        self._type_ctx = runner.get_context(cxx_exposed_type.canonical_pointee_type.unqualified_type_name)

    @cached_property
    def cxx(self):
        return CXXRunnerExposedType(self._cxx_exposed_type._cxx_type,
                                    template_choice=self._cxx_exposed_type._template_choice,
                                    cxx_element=self._cxx_exposed_type._cxx_element,
                                    ir=self._runner.ir,
                                    ctx=self._type_ctx)

    @property
    def vars(self):
        return self._type_ctx.vars if self._type_ctx else types.SimpleNamespace(**self._runner.ir.args)

    @property
    def has_api(self):
        return self._type_ctx is not None

    @property
    def template_args_postfixes(self):
        return self._type_ctx.template_args_postfixes if self._type_ctx else None

    @property
    def template_choice(self):
        return self._type_ctx.template_choice if self._type_ctx else None

    @property
    def descendants(self):
        if self._type_ctx is None or self._type_ctx.node.descendants is None:
            return None

        _descendants = []
        for descendant_node in self._type_ctx.node.descendants:
            if descendant_node.cxx_element.is_templated:
                template_infos = RunRule._get_template_infos(descendant_node)
                for template_info in template_infos:
                    _descendants.append(replace_template_choice(descendant_node.full_displayname,
                                                                template_info.choice))
            else:
                _descendants.append(descendant_node.full_displayname)

        return _descendants



# Copyright (c) 2022 PicsArt, Inc.
# All rights reserved. Use of this source code is governed by a
# MIT-style license that can be found in the LICENSE file.

import os

PROJECT_CONFIG_DIR = os.path.join(os.path.dirname(__file__), "../config/")
YAML_CONFIG_TEMPLATE_PATH = os.path.join(PROJECT_CONFIG_DIR, 'cppbind.yaml')
JINJA_UNIQUE_MARKER = '~!+marker#@~'



# Copyright (c) 2022 PicsArt, Inc.
# All rights reserved. Use of this source code is governed by a
# MIT-style license that can be found in the LICENSE file.

"""
Main module which parses command line arguments and runs cppbind.
"""

import argparse
import cProfile
import json
import os
import sys

import cppbind
from cppbind import default_config, logging, LOG_LEVELS
from cppbind.builder.ir_builder import CppBindIRBuilder
from cppbind.builder.ir_post_processor import IRPostProcessor
from cppbind.builder.out_builder import Builder
from cppbind.common.error import Error, CppBindError
from cppbind.common.yaml_process import to_value
from cppbind.context_manager.ctx_desc import ContextDescriptor
from cppbind.context_manager.ctx_mgr import ContextManager
from cppbind.ir.ast import RootNode
from cppbind.ir.exec_rules import RunRule
from cppbind.parser.cppbind_parser import CXXParser
from cppbind.parser.filter import CppBindFilter
from cppbind.utils import (
    clear_cppbind_generated_files,
    get_host_platform,
    get_var_real_type,
    load_rule_module,
    copy_yaml_config_template,
    absolute_path_from_glob
)


class WrapperGenerator:
    """
    Class for generating wrappers for target language from source files.
    """

    def __init__(self):
        pass

    @staticmethod
    def run(plat_lang_options, ctx_desc, cmd_line_args):
        """Run cppbind for every platform + language combination"""

        logging.info(
            f"Start running wrapper generator for "
            f"{', '.join(list(map(lambda x: x[0] + '.' + x[1], plat_lang_options)))} options.")
        for plat, lang in plat_lang_options:
            WrapperGenerator.run_for(plat, lang, ctx_desc, cmd_line_args)

    @staticmethod
    def run_for(platform, language, ctx_desc, var_values):
        """Run cppbind for current target language + platform"""

        logging.info(f"Start running wrapper generator for "
                     f"{language} language for {platform} platform.")

        ir = RootNode()
        ctx_mgr = ContextManager(ctx_desc, platform, language)
        ir_builder = CppBindIRBuilder(ir, ctx_mgr)

        # Currently `start_root` method is used to evaluate root context and parser method needs root context
        # that's why we call `start_root` here. Maybe later we'll move it into parser logic.
        root_ctx = ir_builder.start_root(var_values)

        if not root_ctx:
            Error.critical(f"""Could not find any config file with path - {default_config.application.context_def_glob}.
                           Run `cppbind init` command under project's root directory to create an initial config file.""")

        logging.debug("Start parsing and building IR.")

        exclude_files = absolute_path_from_glob(root_ctx['src_exclude_glob']) if root_ctx['src_exclude_glob'] else None
        cppbind_filter = CppBindFilter(exclude_files=exclude_files, ir=ir)
        parser = CXXParser(ir_builder, var_values.single_tu, filter_=cppbind_filter)
        parser.parse(**root_ctx)

        # `end_root` is called here to keep symmetry with the `start_root` call
        ir_builder.end_root()

        if Error.has_error():
            Error.critical('Cannot continue: cppbind error has occurred')

        ir_post_processor = IRPostProcessor()
        ir = ir_post_processor.process_ir(ir_builder.ir)
        logging.debug("IR is ready.")

        run_rule = RunRule(ir, ctx_desc, platform, language)
        # load rule modules
        logging.debug("Loading ruler scripts.")
        lang_rule = load_rule_module(language,
                                     default_config.application.rule,
                                     default_config.default_config_dirs)
        logging.debug("Creating builders and running rules on IR.")
        builder = Builder()
        run_rule.run(lang_rule, builder)
        # now we can dump builders into file
        logging.debug("Dumping builders to files.")

        if Error.has_error():
            Error.critical('Cannot continue: cppbind error has occurred')

        builder.dump_outputs()

        if Error.has_error():
            Error.critical('Cannot continue: cppbind error has occurred')


def run(args, ctx_desc):
    """Process language + platform command line arguments and run the main function"""
    plat_lang_options = []
    for option in args.plat_lang_options:
        if '.' in option:
            plat, lang = option.split('.')
        else:
            plat, lang = get_host_platform(), option
        plat_lang_options.append((plat, lang))

    try:
        WrapperGenerator.run(set(plat_lang_options), ctx_desc, args)
    except CppBindError:
        sys.exit(1)


def clean(args):
    """Cleans the files in directory generated by cppbind"""
    if not os.path.isdir(args.dir):
        raise ValueError(f'{args.dir} is not a valid directory.')
    clear_cppbind_generated_files(args.dir)


def init(args):
    """Creates an initial cppbind config file which contains default values."""
    copy_yaml_config_template()


def run_package():
    """
    Command line arguments parser
    """
    profiler = cProfile.Profile()
    # register parent parser
    parent_parser = argparse.ArgumentParser(add_help=False)
    parent_parser.add_argument('--log-level', choices=LOG_LEVELS, type=str, help='Log level', required=False)
    parent_parser.add_argument('--error-limit',
                               type=int,
                               help='Error limit',
                               default=int(default_config.application.error_limit),
                               required=False)
    parent_parser.add_argument('--profile', action='store_true', help='Profiling code and generating reports',
                               required=False)
    parent_parser.add_argument('--report-file',
                               type=str,
                               help='Profiling reports file.',
                               default='profile.prof',
                               required=False)

    parent_parser.add_argument('--single-tu',
                               type=lambda x: (str(x).lower() == 'true'),
                               help='Process all sources as a single translation unit.',
                               default=True)

    parser = argparse.ArgumentParser(description="These are common cppbind commands:")
    sub_parser = parser.add_subparsers(required=True)

    # register clean sub parser
    clean_parser = sub_parser.add_parser('clean', help='Clean all cppbind generated files from directory.',
                                         parents=[parent_parser])
    clean_parser.add_argument('dir', help='Directory from where all cppbind generated files will be deleted.')
    clean_parser.set_defaults(func=clean)

    # register init sub parser
    init_parser = sub_parser.add_parser('init', help='Creates an initial config file in current directory.',
                                        parents=[parent_parser])
    init_parser.set_defaults(func=init)

    # register run sub parser
    run_parser = sub_parser.add_parser('run', help='Run cppbind to generate code for given languages.',
                                       parents=[parent_parser])

    current_sub_parser_args = parent_parser.parse_known_args()

    if current_sub_parser_args[0].profile:
        profiler.enable()
    if current_sub_parser_args and current_sub_parser_args[1][0] == 'run':
        ctx_desc = ContextDescriptor(getattr(default_config.application, 'context_def_glob', None))
        all_languages = ctx_desc.get_deduced_languages()

        plat_lang_choices = all_languages + [plat + '.' + lang for plat in default_config.platforms
                                             for lang in all_languages]
        run_parser.add_argument('plat_lang_options',
                                type=str,
                                nargs='+',
                                choices=plat_lang_choices,
                                help='list of languages for which wrapper will be generated.')

        # add arguments for setting context variables from command line
        for name, prop in ctx_desc.get_var_def().items():
            if 'cmd_line' in prop.get('allowed_on'):
                var_type = get_var_real_type(to_value(prop.get('type')))
                var_desc = to_value(prop.get('description'))

                plat_lang_options = [f"--{plat}.{lang}.{name}" for plat in default_config.platforms
                                     for lang in all_languages]
                plat_options = [f"--{plat}.{name}" for plat in default_config.platforms]
                lang_options = [f"--{lang}.{name}" for lang in all_languages]

                for option in plat_lang_options + plat_options + lang_options + [f"--{name}"]:
                    # suppress option help if it is not pure one (has plat/lang specifications)
                    var_help = var_desc if option == f"--{name}" else argparse.SUPPRESS

                    if var_type is list:
                        run_parser.add_argument(option, help=var_help, nargs='+')
                    elif var_type is dict:
                        run_parser.add_argument(option, help=var_help, type=json.loads)
                    elif var_type is bool:
                        run_parser.add_argument(option, help=var_help, action='store_true')
                    else:
                        run_parser.add_argument(option, help=var_help, type=var_type)

        run_parser.set_defaults(func=lambda arg: run(arg, ctx_desc))

    # print help if nothing is passed
    args = parser.parse_args(args=None if current_sub_parser_args else ['--help'])

    cppbind.init_logger(args.log_level)

    Error.set_error_limit(args.error_limit)

    args.func(args)

    if current_sub_parser_args[0].profile:
        profiler.disable()
        profiler.dump_stats(args.report_file)


if __name__ == "__main__":
    run_package()



# Copyright (c) 2022 PicsArt, Inc.
# All rights reserved. Use of this source code is governed by a
# MIT-style license that can be found in the LICENSE file.


"""Module to define cxx type wrapper class"""

import clang.cindex as cli
import cppbind.utils.clang as cutil


class CXXType:
    """Wrapper type on clang type"""

    def __init__(self, clang_type):
        self.__clang_type = clang_type

    @property
    def type_name(self):
        """The spelling of clang type"""
        return self.__clang_type.spelling

    @property
    def is_function_proto(self):
        """Checks if the type is a function prototype"""
        return self.__clang_type.kind == cli.TypeKind.FUNCTIONPROTO

    @property
    def is_typedef(self):
        """Checks if the type is a typedef to another type"""
        return self.__clang_type.kind == cli.TypeKind.TYPEDEF or \
               self.__clang_type.get_declaration().kind in (cli.CursorKind.TYPE_ALIAS_DECL,
                                                            cli.CursorKind.TYPEDEF_DECL)

    @property
    def is_lval_reference(self):
        """Checks if the type is lvalue reference"""
        return self.__clang_type.kind == cli.TypeKind.LVALUEREFERENCE

    @property
    def is_rval_reference(self):
        """Checks if the type is rvalue reference"""
        return self.__clang_type.kind == cli.TypeKind.RVALUEREFERENCE

    @property
    def is_pointer(self):
        """Checks if the type is a pointer"""
        return self.__clang_type.kind == cli.TypeKind.POINTER

    @property
    def pointee_type(self):
        """Returns the pointee type of the type"""
        return CXXType(self.__clang_type.get_pointee())

    @property
    def canonical_type(self):
        """Returns canonical type of the type"""
        return CXXType(self.__clang_type.get_canonical())

    @property
    def argument_types(self):
        """Returns the list of argument types of function type"""
        return [CXXType(arg_type) for arg_type in self.__clang_type.argument_types()]

    def get_template_declaration(self):
        """Returns the declaration cursor for the clang Type"""
        return cutil.get_template_declaration(self.__clang_type)

    def get_declaration(self):
        """Returns the declaration cursor for the clang Type."""
        return cutil.get_declaration(self.__clang_type)

    def get_result(self):
        """Returns the result type associated with a function type"""
        return CXXType(self.__clang_type.get_result())

    def get_template_argument_type(self, num):
        """Returns the CXXType for the indicated template argument"""
        return CXXType(self.__clang_type.get_template_argument_type(num))

    @property
    def is_enum(self):
        """Checks the type is enum"""
        return self.__clang_type.get_canonical().kind == cli.TypeKind.ENUM

    @property
    def is_bool(self):
        """Checks the type is boolean"""
        return self.__clang_type.get_canonical().kind == cli.TypeKind.BOOL

    @property
    def is_long(self):
        """Checks the type is long"""
        return self.__clang_type.get_canonical().kind == cli.TypeKind.LONG

    @property
    def is_float(self):
        """Checks the type is float"""
        return self.__clang_type.get_canonical().kind in (cli.TypeKind.FLOAT, cli.TypeKind.FLOAT128)

    @property
    def is_char(self):
        """Checks the type is char"""
        return self.__clang_type.get_canonical().kind in (cli.TypeKind.CHAR_S, cli.TypeKind.CHAR_S)

    @property
    def is_uchar(self):
        """Checks the type is unsigned char"""
        return self.__clang_type.get_canonical().kind == cli.TypeKind.UCHAR

    @property
    def base_types(self):
        """Returns a list of base types."""
        _base_types = []
        decl_cursor = self.get_declaration()
        if decl_cursor:
            for base_specifier in decl_cursor.get_children():
                if base_specifier.kind == cli.CursorKind.CXX_BASE_SPECIFIER:
                    _base_types.append(CXXType(base_specifier.type))

        return _base_types

    @property
    def is_variadic(self):
        """Returns whether the type is a variadic template or an instance of a variadic template."""
        if cutil.is_templated(self.__clang_type):
            declaration = self.get_declaration()
            if declaration:
                return any(cutil.is_variadic(cursor) for cursor in
                           cutil.get_template_parameter_cursors(self.get_declaration()))
        return False

    def __getattr__(self, prop):
        """Delegate all missing methods/properties to the underlying clang Type object"""

        if not hasattr(self.__clang_type, prop):
            raise AttributeError(f"{self.__class__.__name__}.{prop} is invalid")

        return getattr(self.__clang_type, prop)



# Copyright (c) 2022 PicsArt, Inc.
# All rights reserved. Use of this source code is governed by a
# MIT-style license that can be found in the LICENSE file.

"""
Processor module provides various processor for cppbind parser
"""
import copy
import datetime
import os
import sys
from collections import OrderedDict
from types import SimpleNamespace

from cppbind import DATETIME_FORMAT, logging, PROJECT_CONFIG_DIR
from cppbind.builder import OUTPUT_MODIFICATION_KEY
from cppbind.common.error import Error
from cppbind.ir.ast import (
    CXXNode,
    DirectoryNode,
    FileNode,
    NodeType,
    RootNode
)
from cppbind.parser.cppbind_api_parser import APIParser
from cppbind.utils import get_android_ndk_sysroot
import clang.cindex as cli
import cppbind.utils.clang as cutil

try:
    import git
except ImportError as e:
    Error.warning("Couldn't find Git. '_get_git_repo_url' system variable will not work properly.")


class CXXPrintProcessor:
    """Class to show an information about a cursor"""
    def __call__(self, cursor, *args, **kwargs):
        print(f'Found {cursor.kind} Display name {cursor.displayname} \
              [line={cursor.location.line}, \
              col={cursor.location.column}]',
              f'Comments {cursor.raw_comment} Brief Comments {cursor.brief_comment}')


class CppBindIRBuilder:
    """
    Class to build intermediate representation.
    """

    def __init__(self, ir, ctx_mgr):
        self.ir = ir
        self.ctx_mgr = ctx_mgr
        self.node_stack = []
        self._sys_vars = {}
        self._processed_dirs = {}
        # cache for holding parent args
        self._parent_arg_mapping = {}
        # list to keep processed file by processing order
        self.processed_api_files = []

    def start_root(self, var_values=None):
        """
        Create root node and eval its context.
        """
        root_node = self.ir
        self.node_stack.append(root_node)
        self.__update_internal_vars(root_node)

        ctx = self.get_full_ctx()
        # update context with initial context provided via command line arguments
        var_values = self.ctx_mgr.filter_by_plat_lang(var_values)

        api, args = self.ctx_mgr.eval_root_attrs(ctx, var_values)
        root_node.api = api
        root_node.args = args

        return args

    def end_root(self):
        assert self.node_stack, "stack should not be empty"
        node = self.node_stack.pop()
        assert node.name == RootNode.ROOT_KEY
        assert len(self.node_stack) == 0, "stack should be empty"
        # set flag which shows that IR built process is completed
        self.ir._finalize_ir_build(self.processed_api_files)

    def start_dir(self, dir_name):
        """
        Create directory node and eval its context.
        """
        if dir_name not in self._processed_dirs:
            file_name = None
            if self.ctx_mgr.ctx_desc.has_yaml_api(dir_name):
                file_name = self.ctx_mgr.ctx_desc.get_api_def_filename(dir_name)

            dir_node = DirectoryNode(dir_name, file_name=file_name, root=self.ir)
            self.node_stack.append(dir_node)
            self._processed_dirs[dir_name] = dir_node

            self.__update_internal_vars(dir_node)
            ctx = self.get_full_ctx()
            if self.ctx_mgr.ctx_desc.has_yaml_api(dir_name):
                # in case dir has a yaml api, yaml node location can be used to report errors properly
                location = None
            else:
                location = SimpleNamespace(file_name=dir_node.file_name,
                                           line_number=dir_node.line_number)
            api, args = self.ctx_mgr.eval_dir_attrs(dir_name, ctx, location)
            dir_node.api = api
            dir_node.args = args
        else:
            # directory is already processed
            dir_node = self._processed_dirs[dir_name]
            self.node_stack.append(dir_node)

    def end_dir(self, dir_name):
        assert self.node_stack, "stack should not be empty"
        node = self.node_stack.pop()
        assert isinstance(node, DirectoryNode), "Should be a directory node"
        assert node.name == dir_name
        # has an API call or child with API call
        if node.api or node.children:
            if len(self.node_stack) > 0:
                if not isinstance(self.node_stack[-1], DirectoryNode):
                    parent_node = self.node_stack[0]
                else:
                    parent_node = self.node_stack[-1]
                if node not in parent_node.children:
                    parent_node.add_child(node)

    def start_file(self, file_name, *args, **kwargs):
        """
        Create file node(can be an include) and eval its context.
        """
        current_node = FileNode(file_name, root=self.ir)
        current_node.args = OrderedDict()
        self.node_stack.append(current_node)
        self.processed_api_files.append(file_name)

        self.__update_internal_vars(current_node)
        ctx = self.get_full_ctx()

        res = self.ctx_mgr.eval_file_attrs(file_name, ctx)
        if res:
            api, args = res
            current_node.api = api
            current_node.args = args

    def end_file(self, file_name):
        """
        Ends current file from node stack and adds it to parent node if it has an api or has children with api.
        """
        tu_node = self.node_stack.pop()
        assert isinstance(tu_node, FileNode), "Should be a file node"
        assert tu_node.spelling == file_name
        if tu_node.api or tu_node.children:  # node has API call or child whit API call
            parent_node = self.node_stack[-1]
            parent_node.add_child(tu_node)
        else:
            # file is not added to IR, remove it from processed list
            self.processed_api_files.remove(file_name)

        # tu is processed it cannot be a parent anymore delete it's args if they're present
        self._parent_arg_mapping.pop(tu_node.full_displayname, None)

    def start_cursor(self, cursor, *args, **kwargs):
        """
        Create a node wrapper for current cursor and eval its context.
        """
        current_node = self.__get_node(cursor)
        self.node_stack.append(current_node)

        cursor_display_name = current_node.full_displayname

        if current_node.api is not None or (not APIParser.has_api(cursor.raw_comment) and
                                            not self.ctx_mgr.ctx_desc.has_yaml_api(cursor_display_name)):
            return

        self.__update_internal_vars(current_node)

        pure_comment = api_section = None
        if APIParser.has_api(cursor.raw_comment):
            pure_comment, api_section = APIParser.separate_pure_and_api_comment(cursor.raw_comment)

        ctx = self.get_full_ctx(pure_comment)
        if self.ctx_mgr.ctx_desc.has_yaml_api(cursor_display_name):
            # in case type has a yaml api, yaml node location can be used to report errors properly
            location = None
        else:
            location = SimpleNamespace(file_name=cursor.extent.start.file.name,
                                       line_number=cursor.extent.start.line)

        res = self.ctx_mgr.eval_clang_attrs(cursor_display_name,
                                            current_node.kind_name,
                                            api_section,
                                            ctx,
                                            location)
        if res:
            api, args = res
            current_node.api = api
            current_node.args = args
            current_node.pure_comment = pure_comment

    def end_cursor(self, cursor, *args, **kwargs):
        node = self.node_stack.pop()
        assert isinstance(node, CXXNode), "Should be a cxx node"
        if node.api or node.children:
            # node has API call or child with API call
            parent_node = self.node_stack[-1]
            parent_node.add_child(node)
        # cursor is processed it cannot be a parent anymore delete it's args if they're present
        self._parent_arg_mapping.pop(node.full_displayname, None)

    def __get_node(self, cursor):
        """Private method to create a new cxx node or return the cached one if it has been created earlier"""
        lookup_name = cutil.get_signature(cursor)
        node = self.ir.find_node(lookup_name)
        if node is not None:
            return node

        return CXXNode(cursor, root=self.ir)

    def get_parent_args(self):
        """
        Get context from the nearest parent which has API.
        """
        if len(self.node_stack) < 2:
            return None

        direct_parent_name = self.node_stack[-2].full_displayname
        if direct_parent_name in self._parent_arg_mapping:
            return self._parent_arg_mapping[direct_parent_name]

        parents = reversed(self.node_stack[:-1])
        parent_args = None
        for parent in parents:
            if parent.api:
                var_def = self.ctx_mgr.ctx_desc.get_var_def()
                # select only those variables which has 'inheritable' property set to true
                # in order not to have redundant values in context when passing to child nodes
                parent_args = {name: val for name, val in parent.args.items()
                               if var_def[name].get('inheritable')}
                break
        self._parent_arg_mapping[direct_parent_name] = parent_args
        return parent_args

    def __update_internal_vars(self, node):
        """
        Update internal variables depending current node type.
        """
        if node.type == NodeType.ROOT_NODE:
            sys_vars = {
                '_output_modification_time': OUTPUT_MODIFICATION_KEY,
                '_current_working_dir': os.getcwd(),
                'path': os.path,
                '_cppbind_extras_dir': PROJECT_CONFIG_DIR,
                'getenv': os.getenv,
                'get_android_ndk_sysroot': get_android_ndk_sysroot,
                '_object_kind': node.kind_name
            }
        else:
            sys_vars = {
                '_output_modification_time': OUTPUT_MODIFICATION_KEY,
                'path': os.path,
                'getenv': os.getenv,
                '_current_working_dir': os.getcwd(),
                '_pure_comment': '',
                '_line_number': node.line_number,
                'source_file_fullname': node.file_name,
                '_object_kind': node.kind_name
            }

        if node.type == NodeType.DIRECTORY_NODE:
            sys_vars.update({
                '_source_modification_time': CppBindIRBuilder._get_modification_time(node.name),
                '_object_name': node.name,
                '_file_name': os.path.splitext(os.path.basename(
                    node.file_name))[0] if node.file_name else node.name,
            })

        elif node.type == NodeType.FILE_NODE:
            # we have a virtual file like extra_includes.cpp so setting mod_time None for these type of files
            mod_time = CppBindIRBuilder._get_modification_time(node.file_name) if os.path.exists(
                node.file_name) else None
            sys_vars.update({
                '_source_modification_time': mod_time,
                '_object_name': node.spelling,
                '_file_name': os.path.splitext(os.path.basename(node.file_name))[0],
            })

        elif node.type == NodeType.CLANG_NODE:
            sys_vars.update({
                '_source_modification_time': CppBindIRBuilder._get_modification_time(node.file_name),
                '_is_operator': node.displayname.startswith('operator'),
                '_object_name': node.spelling,
                '_file_name': os.path.splitext(os.path.basename(node.file_name))[0],
            })

        self._sys_vars.update(sys_vars)

    @staticmethod
    def _get_modification_time(path):
        """Get last modification time of current file"""
        modification_time = datetime.datetime.fromtimestamp(os.stat(path).st_ctime)
        return datetime.date.strftime(modification_time, DATETIME_FORMAT)

    def get_sys_vars(self):
        """
        Get system variables and add git repo url of project directory.
        """
        sys_vars = copy.copy(self._sys_vars)

        def _get_git_repo_url(project_dir):
            if 'git' in sys.modules:
                try:
                    repo = git.Repo(project_dir)
                    url = next(repo.remote().urls).replace('.git', '')
                    branch = repo.active_branch.name
                    logging.info(f'Found a git repo under {project_dir}')
                    return f'{url}/tree/{branch}/'
                except (git.GitError, TypeError):
                    # not a git repo leave variable empty
                    # TypeError may be thrown in case of incorrect reference to a git commit
                    Error.warning(
                        f'Could not find a git repository under: {project_dir}.')
                    return None

        sys_vars['_get_git_repo_url'] = _get_git_repo_url

        return sys_vars

    def get_full_ctx(self, pure_comment=None):
        """
        Construct full context: system variables + pure comment + parent context
        """
        ctx = self.get_sys_vars()
        if pure_comment is None:
            current_node = self.node_stack[-1]
            if current_node.type == NodeType.CLANG_NODE and current_node.cxx_element.raw_comment:
                pure_comment, _ = APIParser.separate_pure_and_api_comment(
                    current_node.cxx_element.raw_comment)
        if pure_comment:
            ctx['_pure_comment'] = '\n'.join(pure_comment)

        parent_args = self.get_parent_args()
        if parent_args:
            ctx.update(parent_args)
        return ctx



# Copyright (c) 2022 PicsArt, Inc.
# All rights reserved. Use of this source code is governed by a
# MIT-style license that can be found in the LICENSE file.

"""
Builder module allows rule script to add source code blokes
to generated output.
"""
import copy
import os

from cppbind import logging, current_datetime, Error
from cppbind.builder import is_output_changed, OUTPUT_MODIFICATION_KEY

TAB_STR = '    '


class Scope:

    def __init__(self, *parts, name=None, tab=0, tab_str=None, file_scope=None, parts_splitter='\n'):
        self.file_scope = file_scope
        self.parts = []
        self.tab = tab
        self.parts_splitter = parts_splitter
        self.tab_str = tab_str or TAB_STR
        self._name = name
        self.add(*parts)

    @property
    def name(self):
        return self._name

    def add_unique(self, *parts):
        """
        Adds text to corresponding partition
        """
        return self.add(
            *[part for part in list(dict.fromkeys(parts))
              if part not in self.parts])

    def add(self, *parts):
        """
        Adds text or scope to corresponding partition
        """
        self.parts = self.parts + list(parts)

        # register scopes
        self._register_scopes(*parts)

    def get_scope(self, scope_name, create=False, add_as_part=True):
        scope = self.file_scope.lookup_scope(scope_name)

        if scope is None and create:
            logging.debug(f"Creating Scope {scope_name},"
                          f"current depth is {len(self.file_scope._scope_stack)}.")
            scope = Scope(name=scope_name, file_scope=self.file_scope)
            if add_as_part:
                self.add(scope)
            else:
                self._register_scopes(scope)

        return scope

    def create_scope(self, scope_name):
        logging.debug(f"Creating Scope {scope_name}, "
                      f"current depth is {len(self.file_scope._scope_stack)}.")
        scope = Scope(name=scope_name, file_scope=self.file_scope)
        self._register_scopes(scope)
        return scope

    def _register_scopes(self, *parts):
        if len(parts) == 0:
            return
        # register scopes
        if self.file_scope is not None:
            if self is not self.file_scope:
                dept = self.file_scope._lookup_scope_dept(self)
                assert dept is not None, f"current {repr(self)} scope is not register"
                # register as child
                if dept != -1:
                    dept += 1
            else:
                dept = 0

        for data in parts:
            if data is not None:
                if isinstance(data, Scope):
                    data.file_scope = self.file_scope
                    # also if name is not empty add to file_scope for lookup
                    if data.name:
                        assert self.file_scope is not None, \
                            "to be able to add name scope file_scope should be specified."
                        self.file_scope.register_scope(data, dept)

    def __getitem__(self, scope_name):
        # This is not called for now, and value of the create might be incorrect.
        return self.get_scope(scope_name, create=True)

    def __repr__(self):
        return f"Scope({self.__dict__})"

    def __str__(self):

        # `if str(p)` helps to skip blank lines
        lines = [
            (self.tab_str * self.tab) + s
            for s in self.parts_splitter.join(str(p) for p in self.parts if str(p)).splitlines(keepends=True)
        ]

        str_scope = ''.join(lines)

        return str_scope

    def __bool__(self):
        # this check helps to remove redundant blank lines
        return any(str(p) for p in self.parts)


class File(Scope):

    def __init__(self, file_path, **kwargs):
        self.file_path = file_path
        self._scope_stack = []
        super().__init__(file_scope=self, **kwargs)

    def dump_output(self):
        logging.info(f"Writing output for {self.name} into {self.file_path}")
        os.makedirs(os.path.dirname(self.file_path), exist_ok=True)
        content = str(self)
        try:
            if os.path.exists(self.file_path):
                with open(self.file_path, 'r+') as f:
                    old = f.read()
                    if is_output_changed(old, content):
                        f.seek(0)
                        f.truncate()
                        content = content.replace(OUTPUT_MODIFICATION_KEY,
                                                  current_datetime())
                        f.write(content)
            else:
                with open(self.file_path, 'w') as f:
                    content = content.replace(OUTPUT_MODIFICATION_KEY,
                                              current_datetime())
                    f.write(content)
        except OSError as err:
            Error.critical(f"Cannot do I/O operation with {self.file_path} file: {err}")

    def register_scope(self, scope, dept=-1):
        assert self.scope_stack and scope.name
        self.scope_stack[dept].setdefault(scope.name, scope)

    def add_scope_stack(self):
        self._scope_stack.append(dict())

    def pop_scope_stack(self):
        self._scope_stack.pop()

    def clear_scope_stack(self):
        self.scope_stack[-1] = {}

    def lookup_scope(self, scope_name):
        scope, _ = self._lookup_scope_by_name(scope_name)
        return scope

    @property
    def scope_stack(self):
        return self._scope_stack

    def __getitem__(self, scope_name):
        return self.get_scope(scope_name, create=False)

    def _lookup_scope_by_name(self, scope_name, dept=-1):
        if len(self.scope_stack) < -dept:
            return None, -1

        scope = self.scope_stack[dept].get(scope_name, None)
        if scope is not None:
            return scope, dept

        return self._lookup_scope_by_name(scope_name, dept - 1)

    def _lookup_scope_dept(self, scope, dept=-1):
        if len(self.scope_stack) < -dept:
            return None

        found_scope, found_dept = self._lookup_scope_by_name(scope.name, dept)
        if found_scope is None:
            return None

        if found_scope is scope:
            return found_dept
        return self._lookup_scope_dept(scope, found_dept - 1)


class Builder:

    def __init__(self):
        self._current_dept = 0
        self._files = {}

    def dump_outputs(self):
        for _, fl in self._files.items():
            fl.dump_output()

    def get_file(self, file_name, file_path=None, create=True, init_func=None):
        assert init_func is None or create, "init func should specified only if create is true."
        file_path = file_path or file_name
        file_scope = self._files.get(file_path, None)
        assert file_scope is None or file_scope.name == file_name

        if file_scope is None and create:
            logging.debug(f"Creating File Scope {file_path}, "
                          f"current depth is {self._current_dept}.")
            file_scope = File(file_path, name=file_name)
            for _ in range(self._current_dept):
                file_scope.add_scope_stack()
            self._files[file_path] = file_scope
            if init_func:
                init_func(file_scope)

        return file_scope

    def add_scope_stack(self):
        self._current_dept += 1
        for fl in self._files.values():
            fl.add_scope_stack()
            assert len(fl._scope_stack) == self._current_dept, \
                f"dept imbalance at {self._current_dept}, got {len(fl._scope_stack)} for file {fl.name}"

    def pop_scope_stack(self):
        self._current_dept -= 1
        for fl in self._files.values():
            fl.pop_scope_stack()
            assert len(fl._scope_stack) == self._current_dept, \
                f"dept imbalance at {self._current_dept}, got {len(fl._scope_stack)} for file {fl.name}"

    def clear_scope_stack(self):
        for fl in self._files.values():
            fl.clear_scope_stack()

    def capture_stacks(self):
        return {file_name: copy.copy(fl.scope_stack) for file_name, fl in self._files.items()}

    def restore_stacks(self, capture_data):
        restoring_dept = None
        for file_name, fl in self._files.items():
            if file_name in capture_data:
                fl._scope_stack = copy.copy(capture_data[file_name])
                if restoring_dept is None:
                    restoring_dept = len(fl._scope_stack)
                else:
                    assert restoring_dept == len(fl._scope_stack), \
                        f"Scope imbalance for {file_name} scope dept is {len(fl._scope_stack)}"
            else:
                logging.debug(f"Current scope is {self._current_dept}, "
                              f"{file_name} scope is {len(fl._scope_stack)}.")

        if restoring_dept is not None:
            self._current_dept = restoring_dept

        for file_name, fl in self._files.items():
            s_len = len(fl._scope_stack)
            fl._scope_stack += [{}] * (self._current_dept - s_len)



# Copyright (c) 2022 PicsArt, Inc.
# All rights reserved. Use of this source code is governed by a
# MIT-style license that can be found in the LICENSE file.

"""
Helper functions working with clang
"""
import os
import re
import sys
from ctypes import *
from ctypes.util import find_library
from itertools import chain

from cachetools import cached, LRUCache

import clang.cindex as cli
from cppbind.common.error import Error


def template_type_name(type_):
    name = get_unqualified_type_name(type_)
    end_idx = name.find('<')
    if end_idx != -1:
        return name[:end_idx].strip()
    return name


def is_overloaded(cursor):
    return bool([child for child in list(cursor.semantic_parent.get_children()) if
                 child.spelling == cursor.spelling and child != cursor])


def _get_unqualified_type_name(type_name):
    """
    TODO: python API is missing
    """
    qualifiers = ["const ", "volatile "]
    for qual in qualifiers:
        if type_name.startswith(qual):
            return _get_unqualified_type_name(type_name[len(qual):])
    return type_name


def get_unqualified_type_name(type_):
    return _get_unqualified_type_name(type_.spelling if isinstance(type_, cli.Type) else type_)


def get_semantic_ancestors(cursor):
    ancestors = []
    _cursor = cursor.semantic_parent

    while _cursor:
        ancestors.append(_cursor)
        _cursor = _cursor.semantic_parent

    return ancestors[::-1]


def join_type_parts(type_parts):
    return '::'.join(type_parts)


def get_full_displayname(cursor):
    ancestors = get_semantic_ancestors(cursor)
    spellings = [a.displayname for a in ancestors[1::] if a.displayname != '']
    spellings.append(cursor.displayname)
    return join_type_parts(spellings)


def get_signature(cursor):
    full_displayname = get_full_displayname(cursor)
    if cursor.kind == cli.CursorKind.NAMESPACE:
        # namespace cursors coming from #include directives has relative file name
        return join_type_parts([os.path.abspath(cursor.extent.start.file.name),
                                str(cursor.extent.start.line),
                                full_displayname])
    return full_displayname


def is_final_cursor(cursor):
    if cursor.kind == cli.CursorKind.CXX_METHOD:
        if not cursor.is_virtual_method():
            return True
        return cli.CursorKind.CXX_FINAL_ATTR in chain(
            (c.kind for c in cursor.get_children()),
            (c.kind for c in cursor.semantic_parent.get_children())
        )
    return cli.CursorKind.CXX_FINAL_ATTR in (c.kind for c in cursor.get_children())


@cached(cache=LRUCache(maxsize=128), key=lambda cursor: get_signature(cursor))
def is_polymorphic(cursor):
    """
    Returns true if the class/struct is polymorphic (contains or inherits a virtual function)
    Note: There is `isPolymorphic()` class in libclang `clang::CXXRecordDecl` class, but it is not exported
    """
    if cursor.kind not in (cli.CursorKind.STRUCT_DECL,
                           cli.CursorKind.CLASS_DECL,
                           cli.CursorKind.CLASS_TEMPLATE,
                           cli.CursorKind.CLASS_TEMPLATE_PARTIAL_SPECIALIZATION):
        Error.internal("Is polymorphic check is valid only for class/struct types")

    base_cursors = []
    for child in cursor.get_children():
        if child.kind in (cli.CursorKind.CXX_METHOD,
                          cli.CursorKind.DESTRUCTOR) and child.is_virtual_method():
            return True

        # collect all bases to process later (no need to dive into bases in case the type itself is virtual)
        if child.kind == cli.CursorKind.CXX_BASE_SPECIFIER:
            base_cursors.append(child.referenced)

    return any(is_polymorphic(base_cursor) for base_cursor in base_cursors)


@cached(cache=LRUCache(maxsize=128), key=lambda cursor: get_signature(cursor))
def has_multiple_base_branches(cursor):
    """Returns true if the class/struct has multiple base branches in its hierarchy"""
    if cursor.kind not in (cli.CursorKind.STRUCT_DECL,
                           cli.CursorKind.CLASS_DECL,
                           cli.CursorKind.CLASS_TEMPLATE,
                           cli.CursorKind.CLASS_TEMPLATE_PARTIAL_SPECIALIZATION):
        Error.internal("Has multiple bases check is valid only for class/struct types")

    base_cursors = [child for child in cursor.get_children() if child.kind == cli.CursorKind.CXX_BASE_SPECIFIER]
    if len(base_cursors) > 1:
        return True

    return has_multiple_base_branches(base_cursors[0].referenced) if base_cursors else False


def extract_pure_comment(raw_comment, end_index=None):
    """
    Returns pure comment(without API part).
    Args:
        raw_comment(str): Doxygen comment.
        end_index(int): Index to where to consider,
                        if not specified then will consider the whole string.
    Returns:
        str: Characters removed type_name.
    """
    end_index = end_index or len(raw_comment)
    # using one lstrip(' ') before removing any of the given characters
    # and one at the end to remove all leading whitespaces
    comment_lines = [(comment_line.lstrip(' ')).lstrip('/*!').lstrip(' ')
                     for comment_line in raw_comment[:end_index].splitlines()]

    if comment_lines and not comment_lines[0]:
        comment_lines.pop(0)
    if comment_lines and not comment_lines[-1]:
        comment_lines.pop()

    return comment_lines


def is_templated(type_):
    """
    Check whether cursor/type is templated (is a template or depends on a template parameter),
    or string corresponds to a template cursor signature
    """

    if isinstance(type_, cli.Type):
        return type_.get_num_template_arguments() != -1
    elif isinstance(type_, cli.Cursor):
        is_parent_template = is_templated(type_.lexical_parent) if type_.lexical_parent else False
        return type_.kind in (cli.CursorKind.CLASS_TEMPLATE,
                              cli.CursorKind.CLASS_TEMPLATE_PARTIAL_SPECIALIZATION,
                              cli.CursorKind.FUNCTION_TEMPLATE) or is_parent_template

    return '<' in type_ and '>' in type_


def replace_template_choice(type_name, template_choice):
    """
    Return type name with replaced template arguments,
    e.g. for a::Foo<T,V> will return a::Foo<Project,int>.
    Args:
        type_name(str): Type name. e.g. a::Foo<T,V>
        template_choice (dict): Containing template current value,
                                e.g. {"T": "a::Project", "V": "int"}
    Returns:
        str: Replaced type_name.
    """
    replaced = type_name
    if template_choice:
        if replaced in template_choice:
            return template_choice[replaced]
        for typename, value in template_choice.items():
            regex = rf'(^|[,<\s])(\s*){typename}([\s,>&.*]|$)'
            val = ', '.join(value) if isinstance(value, list) else value
            if val == '':
                # val is an empty string when variadic template is used with an empty argument list e.g., tuple<>
                match = re.search(regex, replaced)
                # check if the first argument is empty and do not replace '<'
                # e.g. for tuple<Ts...> will be tuple<...>
                # for tuple<int, Ts...> will be tuple<int...>
                g1 = '<' if match and '<' in match.group(1) else ''
                g3 = '' if g1 else '\g<3>'
                g3 = '>' if match and '>' in match.group(3) else g3
                replaced = re.sub(regex, rf'{g1}{val}{g3}', replaced)
            else:
                replaced = re.sub(regex, rf'\g<1>\g<2>{val}\g<3>', replaced)
        replaced = replaced.replace('.', '')
    return replaced


def is_variadic(cursor):
    """
    Returns whether the cursor represents a template parameter pack or not.
    """
    return any(t.spelling == '...' for t in cursor.get_tokens())


def is_integral_type(clang_type):
    """
    Returns whether clang Type is of integral type or not.
    """
    return clang_type.kind in (cli.TypeKind.SHORT,
                               cli.TypeKind.USHORT,
                               cli.TypeKind.INT,
                               cli.TypeKind.UINT,
                               cli.TypeKind.LONG,
                               cli.TypeKind.ULONG,
                               cli.TypeKind.LONGLONG,
                               cli.TypeKind.ULONGLONG,
                               cli.TypeKind.INT128,
                               cli.TypeKind.UINT128)


# TODO consistent-cxx-info: this logic is related to the clang, should go to IR after refactoring
def get_template_declaration(clang_type):
    """
    Returns the declaration cursor for the template clang Type and None for non-template types.
    Args:
        clang_type (clang.cindex.Type):
    Returns:
        clang.cindex.Cursor: Declaration cursor.
    """
    if not is_templated(clang_type):
        return None
    decl_cursor = clang_type.get_declaration()
    if decl_cursor.kind == cli.CursorKind.CLASS_TEMPLATE:
        return decl_cursor
    spec_cursor = clang_type.get_declaration().get_specialized_cursor_template()
    # the declaration is a specialized type cursor
    while spec_cursor:
        if spec_cursor.kind == cli.CursorKind.CLASS_TEMPLATE:
            return spec_cursor
        spec_cursor = spec_cursor.get_specialized_cursor_template()
    pointee = clang_type.get_pointee()
    if pointee.kind != cli.TypeKind.INVALID:
        return get_template_declaration(pointee)
    return get_template_declaration(clang_type.get_canonical())


def get_declaration(clang_type):
    """
    Returns the declaration cursor for the clang Type.
    Args:
        clang_type (clang.cindex.Type):
    Returns:
        clang.cindex.Cursor: Declaration cursor.
    """
    if is_templated(clang_type):
        return get_template_declaration(clang_type)
    return _get_declaration(clang_type)


def _get_declaration(clang_type):
    if clang_type.kind in (cli.TypeKind.INVALID, cli.TypeKind.UNEXPOSED):
        return None
    pointee = clang_type.get_pointee()
    if pointee.kind != cli.TypeKind.INVALID:
        return _get_declaration(pointee)
    decl_cursor = clang_type.get_declaration()
    if decl_cursor.kind == cli.CursorKind.NO_DECL_FOUND:
        return None
    if decl_cursor.kind in (cli.CursorKind.CLASS_DECL, cli.CursorKind.STRUCT_DECL, cli.CursorKind.ENUM_DECL):
        return decl_cursor
    return _get_declaration(clang_type.get_canonical())


# TODO consistent-cxx-info: this logic is related to the clang, should go to IR after refactoring
def get_template_parameter_cursors(clang_cursor):
    """
    Returns the list of template parameter cursors for the clang cursor.
    Args:
        clang_cursor (clang.cindex.Cursor):
    Returns:
        list[clang.cindex.Cursor]: List of argument cursors.
    """
    return [child for child in clang_cursor.get_children() if child.kind in [
        cli.CursorKind.TEMPLATE_TYPE_PARAMETER,
        cli.CursorKind.TEMPLATE_NON_TYPE_PARAMETER,
        cli.CursorKind.TEMPLATE_TEMPLATE_PARAMETER,
    ]]


# TODO consistent-cxx-info: this logic is related to the clang, should go to IR after refactoring
def get_template_arguments_from_str(type_spelling):
    """
    Retrieves template argument spellings from a type spelling.
    E.g. for 'std::pair<std::string, std::vector<int>>' returns ['std::string', 'std::vector<int>']
    Note: there are cases this function does not handle for example if the expression contains offset operator
    Args:
        type_spelling (str):
    Returns:
        list[str]: List of arguments.
    """
    # only most nested type's arguments should be retrieved e.g for for a::b::<c::C>::G<d::D, e::E> -- d::D, e::E
    template_args = []
    parentheses_count = 0
    length = len(type_spelling)
    arg_end_idx = length - 1
    for i, symbol in enumerate(reversed(type_spelling)):
        # template should end with >
        if symbol == '>':
            parentheses_count += 1
        elif symbol == '<':
            parentheses_count -= 1
        elif symbol == ',' and parentheses_count == 1:
            template_args.append(type_spelling[length - i + 1:arg_end_idx].strip())
            arg_end_idx = length - i - 1
        if parentheses_count == 0:
            # if i=0 then self is not a template
            arg = type_spelling[length - i:arg_end_idx].strip()
            # consider case with empty arguments, e.g. tuple<>
            if i != 0 and arg:
                template_args.append(type_spelling[length - i:arg_end_idx].strip())
            break
    return list(reversed(template_args))


def get_libclang_full_path():
    """
    Return libclang full path.
    cindex.Config.library_file returns only file name for linux not a full path.
    """
    # linkmap structure, we only need the second entry
    if sys.platform.startswith('linux'):
        class LINKMAP(Structure):
            _fields_ = [
                ('l_addr', c_void_p),
                ('l_name', c_char_p)
            ]

        libc = CDLL(cli.Config.library_file)
        libdl = CDLL(find_library('dl'))

        dlinfo = libdl.dlinfo
        dlinfo.argtypes = c_void_p, c_int, c_void_p
        dlinfo.restype = c_int

        # gets typecasted later, I dont know how to create a ctypes struct pointer instance
        lmptr = c_void_p()

        # 2 equals RTLD_DI_LINKMAP, pass pointer by reference
        dlinfo(libc._handle, 2, byref(lmptr))

        # typecast to a linkmap pointer and retrieve the name.
        abspath = cast(lmptr, POINTER(LINKMAP)).contents.l_name
        return abspath.decode('utf-8')
    return cli.Config.library_file



# Copyright (c) 2022 PicsArt, Inc.
# All rights reserved. Use of this source code is governed by a
# MIT-style license that can be found in the LICENSE file.

from .cxx_exposed_element import (
    CXXArgumentExposedElement,
    CXXClassExposedElement,
    CXXEnumExposedElement,
    CXXExposedElement,
    CXXFunctionExposedElement,
    CXXMemberExposedElement
)
from .cxx_exposed_type import CXXExposedType, CXXRunnerExposedType


# Copyright (c) 2022 PicsArt, Inc.
# All rights reserved. Use of this source code is governed by a
# MIT-style license that can be found in the LICENSE file.

"""
Module is responsible for yaml config files loading, merging and context variables forming.
"""

import copy
import glob
import os
from collections import defaultdict
from itertools import chain

import yaml

from cppbind import default_config
from cppbind.common.error import Error
from cppbind.common.yaml_process import load_yaml, YamlKeyDuplicationError
from cppbind.ir.ast import RootNode
from cppbind.utils.clang import join_type_parts

NODE_GROUP_ALIASES = {
    'cmd_line': ('cmd_line', 'root'),
    'file_system': ('dir', 'file'),
    'cxx_method': ('method', 'method_template'),
    'cxx_function': ('function', 'function_template'),
    'cxx_constructor': ('constructor', 'constructor_template'),
    'cxx': (
        'class', 'class_template', 'class_template_partial_specialization', 'struct', 'struct_template', 'constructor', 'constructor_template',
        'function', 'function_template', 'method', 'method_template', 'enum', 'field'
    ),
}


class ContextDescriptor:
    """
    A class to read/load/merge yaml config files and calc/eval context variables
    """

    ROOT_SECTION_KEY = 'vars'
    DIR_SECTION_KEY = 'dir_vars'
    FILE_SECTION_KEY = 'file_vars'
    TYPE_SECTION_KEY = 'type_vars'
    VAR_DEF_SECTION_KEY = 'var_def'

    RULE_SECTION_KEY = 'rules'
    CODE_SNIPPETS_KEY = 'code_snippets'
    TYPE_CONVERTER_SNIPPETS_KEY = 'type_converters'
    ACTION_SNIPPETS_KEY = 'actions'

    VARS_RULE_KEY = 'vars'
    TYPE_RULE_KEY = 'type'
    DIR_RULE_KEY = 'dir'
    FILE_RULE_KEY = 'file'
    SUB_RULE_KEY = ':'

    def __init__(self, context_def_glob):
        self.__ctx_def_map = ContextDescriptor.build_ctx_def_map(context_def_glob)
        self.__deduced_languages = self.__validate_and_deduce_languages()
        self.__var_def = ContextDescriptor.resolve_attr_aliases(self.__get_raw_var_def())
        self.__validate_var_def()

    @staticmethod
    def resolve_attr_aliases(attrs):
        """
        A function to replace node group aliases with their actual values list
        """

        for field in ['allowed_on', 'required_on']:
            for key, val in attrs.items():
                if field not in val:
                    attrs[key][field] = set(chain(*NODE_GROUP_ALIASES.values()) if field == 'allowed_on' else [])
                else:
                    res = set()
                    for node in val[field]:
                        if node in NODE_GROUP_ALIASES:
                            res.update(NODE_GROUP_ALIASES[node])
                        else:
                            res.add(node)
                    attrs[key][field] = res

        return attrs

    @classmethod
    def build_ctx_def_map(cls, context_def_glob):
        """
        A method for iterating over yaml config files and
        building context definition map by loading/merging yaml files.
        """
        if context_def_glob is None:
            return {}

        files = set()
        for file in context_def_glob.split(','):
            files_glob = glob.glob(file.strip(), recursive=True)
            for file_path in files_glob:
                files.add(os.path.abspath(file_path))

        ctx_def_map = {}
        for current_file in list(sorted(files)):
            try:
                attrs = load_yaml(current_file)
            except yaml.YAMLError as err:
                raise yaml.YAMLError(f"Wrong yaml format: {current_file}: {err}")

            if attrs:
                cls.load_merge_ctx_def_map(attrs, ctx_def_map)

        return ctx_def_map

    @classmethod
    def load_merge_ctx_def_map(cls, attrs, ctx_def_map):
        """
        A method for loading/merging yaml config files corresponding sections.
        """
        def flatten_dict(src_dict, section_key, ancestors):
            cls.validate_section_keys(src_dict, section_key)
            if cls.TYPE_RULE_KEY in src_dict:
                ancestors.append(src_dict[cls.TYPE_RULE_KEY].value)
            try:
                if cls.VARS_RULE_KEY in src_dict:
                    flat_key = ContextDescriptor._get_key(src_dict=src_dict,
                                                          ancestors=ancestors)
                    if flat_key in ctx_def_map:
                        cls._raise_redefinition_error(flat_key, src_dict[cls.VARS_RULE_KEY], ctx_def_map[flat_key])
                    ctx_def_map[flat_key] = src_dict[cls.VARS_RULE_KEY]
                if cls.SUB_RULE_KEY in src_dict:
                    for sub in src_dict[cls.SUB_RULE_KEY]:
                        flatten_dict(sub, section_key, ancestors)
            finally:
                if cls.TYPE_RULE_KEY in src_dict:
                    ancestors.pop()

        if cls.ROOT_SECTION_KEY in attrs:
            root_section = attrs[cls.ROOT_SECTION_KEY]
            if RootNode.ROOT_KEY in ctx_def_map:
                cls._raise_redefinition_error(cls.ROOT_SECTION_KEY, root_section, ctx_def_map[RootNode.ROOT_KEY])
            ctx_def_map[RootNode.ROOT_KEY] = attrs[cls.ROOT_SECTION_KEY]

        for section in (cls.DIR_SECTION_KEY, cls.FILE_SECTION_KEY, cls.TYPE_SECTION_KEY):
            if section in attrs:
                if attrs[section].value is not None:
                    for item in attrs[section]:
                        flatten_dict(item, section, [])

        if cls.VAR_DEF_SECTION_KEY in attrs:
            var_def_section = attrs[cls.VAR_DEF_SECTION_KEY]
            if cls.VAR_DEF_SECTION_KEY in ctx_def_map:
                intersection = var_def_section.keys() & ctx_def_map[cls.VAR_DEF_SECTION_KEY].keys()
                if intersection:
                    redefinitions = '\n'.join([f"{redefined_var}: {var_def_section[redefined_var].line_number} - "
                                               f"{ctx_def_map[cls.VAR_DEF_SECTION_KEY][redefined_var].line_number}"
                                               for redefined_var in intersection])
                    raise YamlKeyDuplicationError(f"Variable definition duplicate(s) in\n"
                                                  f"{ctx_def_map[cls.VAR_DEF_SECTION_KEY].file}"
                                                  f" - {var_def_section.file}: \n{redefinitions}.")
                ctx_def_map[cls.VAR_DEF_SECTION_KEY].update(var_def_section)
            else:
                ctx_def_map[cls.VAR_DEF_SECTION_KEY] = var_def_section

        cls.load_merge_rules(attrs, ctx_def_map)

    @classmethod
    def _raise_redefinition_error(cls, key, current_section, previous_section):
        raise YamlKeyDuplicationError(f"Redefinition of '{key}' "
                                      f"section in line {current_section.line_number} "
                                      f"of {current_section.file} file, which must be "
                                      f"uniquely specified only in one file.\n"
                                      f"It was previously defined in line "
                                      f"{previous_section.line_number} "
                                      f"of {previous_section.file} file.")

    @classmethod
    def _get_key(cls, src_dict, ancestors):
        """
        A method to construct a key from loaded yaml config file's current section.
        """

        # dir
        if cls.DIR_RULE_KEY in src_dict:
            dir_name = src_dict[cls.DIR_RULE_KEY].value
            if os.path.isabs(dir_name):
                # if an absolute path is specified then we assume it's absolute to current dir
                return dir_name.replace('/', '', 1) or '.'
            return os.path.relpath(
                os.path.abspath(os.path.join(os.path.dirname(
                    src_dict[cls.DIR_RULE_KEY].file), dir_name)),
                os.getcwd())
        # file
        if cls.FILE_RULE_KEY in src_dict:
            file_name = src_dict[cls.FILE_RULE_KEY].value
            if os.path.isabs(file_name):
                # if an absolute path is specified then we assume it's absolute to current dir
                flat_key = file_name.replace('/', '', 1)
                return os.path.join(flat_key, os.getcwd())
            return os.path.abspath(os.path.join(os.path.dirname(
                src_dict[cls.FILE_RULE_KEY].file), file_name))

        # type
        return join_type_parts(ancestors)

    @staticmethod
    def load_merge_rules(attrs, ctx_def_map):
        """
        A method for loading/merging snippets related sections from yaml config files.
        """

        def merge_rules(rules, rules_map, path):
            """
            Method to merge current tree with global one and report errors
            """
            # actions are list items
            if isinstance(rules_map, list):
                rules_map.extend(rules)
                return
            for key, val in rules.items():
                # keep current key path (list of parent keys) to use in error message
                path.append(key)
                if key in rules_map:
                    if val.is_of_type(dict) and rules_map[key].is_of_type(dict):
                        merge_rules(val, rules_map[key], path)
                    else:
                        raise YamlKeyDuplicationError(f"Redefinition of '{'.'.join(path)}' "
                                                      f"key section in line {val.line_number} "
                                                      f"of {val.file} file,\n"
                                                      f"which already has been defined in line "
                                                      f"{rules_map[key].line_number} of "
                                                      f"{rules_map[key].file} file")
                else:
                    rules_map[key] = copy.deepcopy(val)
                path.pop()

        current_rules = attrs.get(ContextDescriptor.RULE_SECTION_KEY)
        if current_rules:
            for rule_key in current_rules:
                platforms = lang = section_type = None
                rule_key_tokens = rule_key.split('.')

                if len(rule_key_tokens) == 3:
                    platforms = [rule_key_tokens[0]]
                    lang = rule_key_tokens[1]
                    section_type = rule_key_tokens[2]
                elif len(rule_key_tokens) == 2:
                    platforms = default_config.platforms
                    lang = rule_key_tokens[0]
                    section_type = rule_key_tokens[1]

                if lang is None:
                    continue

                for plat in platforms:
                    if section_type == ContextDescriptor.CODE_SNIPPETS_KEY:
                        global_code_snippets = ctx_def_map.setdefault(ContextDescriptor.CODE_SNIPPETS_KEY,
                                                                      defaultdict(lambda: defaultdict(dict)))
                        merge_rules(current_rules[rule_key], global_code_snippets[lang][plat], [])
                    elif section_type == ContextDescriptor.TYPE_CONVERTER_SNIPPETS_KEY:
                        global_type_converters = ctx_def_map.setdefault(ContextDescriptor.TYPE_CONVERTER_SNIPPETS_KEY,
                                                                        defaultdict(lambda: defaultdict(dict)))
                        merge_rules(current_rules[rule_key], global_type_converters[lang][plat], [])
                    elif section_type == ContextDescriptor.ACTION_SNIPPETS_KEY:
                        global_actions = ctx_def_map.setdefault(ContextDescriptor.ACTION_SNIPPETS_KEY,
                                                                defaultdict(lambda: defaultdict(list)))
                        merge_rules(current_rules[rule_key], global_actions[lang][plat], [])

    @classmethod
    def validate_section_keys(cls, src_dict, section_key):
        """
        A method to validate current section and raise an error
        in case of disallowed key usages in the section.
        """
        if section_key == cls.FILE_SECTION_KEY and (cls.DIR_RULE_KEY in src_dict or cls.TYPE_RULE_KEY in src_dict):
            Error.critical(f"API file section definition can only contain '{cls.FILE_RULE_KEY}' key")
        elif section_key == cls.DIR_SECTION_KEY and (cls.FILE_RULE_KEY in src_dict or cls.TYPE_RULE_KEY in src_dict):
            Error.critical(f"API dir section definition can only contain '{cls.DIR_RULE_KEY}' key")
        elif section_key == cls.TYPE_SECTION_KEY and (cls.DIR_RULE_KEY in src_dict or cls.FILE_RULE_KEY in src_dict):
            Error.critical(f"API type section definition can only contain '{cls.TYPE_RULE_KEY}' key")

    def __validate_var_def(self):
        """Method for variables definition validation"""
        for var_name, prop in self.get_var_def().items():
            extra_nodes = set(prop['required_on']) - set(prop['allowed_on'])
            if extra_nodes:
                Error.error(f"Variable '{var_name}' is required on '{', '.join(extra_nodes)}' "
                            f"node{'s' if len(extra_nodes) > 1 else ''}, while it's not allowed on "
                            f"{'them' if len(extra_nodes) > 1 else 'it'}",
                            prop.file,
                            prop.line_number)

    def __validate_and_deduce_languages(self):
        if ContextDescriptor.CODE_SNIPPETS_KEY not in self.__ctx_def_map:
            Error.critical("CppBind error: there is no any specified code snippet rule")
        if ContextDescriptor.TYPE_CONVERTER_SNIPPETS_KEY not in self.__ctx_def_map:
            Error.critical("CppBind error: there is no any specified type converter snippet rule")

        code_snippets_languages = sorted(list(self.__ctx_def_map[ContextDescriptor.CODE_SNIPPETS_KEY].keys()))
        type_converter_languages = sorted(
            list(self.__ctx_def_map[ContextDescriptor.TYPE_CONVERTER_SNIPPETS_KEY].keys()))
        if code_snippets_languages != type_converter_languages:
            languages_with_diff = set(code_snippets_languages).symmetric_difference(set(type_converter_languages))
            Error.critical(f"CppBind error: code snippets and type converters must be both specified for a language. "
                           f"This rule is violated for {', '.join(languages_with_diff)} "
                           f"language{'s' if len(languages_with_diff) > 1 else ''}")

        return code_snippets_languages

    def __get_raw_var_def(self):
        """Get raw variable definitions section without resolving node alias names"""
        return self.__ctx_def_map.get(ContextDescriptor.VAR_DEF_SECTION_KEY, {})

    def get_var_def(self):
        """Get variable definitions with already resolved node alias names"""
        return self.__var_def

    def get_code_snippets(self):
        """Get final code snippets section"""
        return self.__ctx_def_map.get(ContextDescriptor.CODE_SNIPPETS_KEY, {})

    def get_type_converter_snippets(self):
        """Get type converter snippets section"""
        return self.__ctx_def_map.get(ContextDescriptor.TYPE_CONVERTER_SNIPPETS_KEY, {})

    def get_action_snippets(self):
        """Get action snippets section"""
        return self.__ctx_def_map.get(ContextDescriptor.ACTION_SNIPPETS_KEY, [])

    def has_yaml_api(self, name):
        """
        Check whether current name is present in context definition map
        """
        return name in self.__ctx_def_map

    def get_api_def_filename(self, name):
        """
        Method to get yaml config file name in which file/dir api is defined
        """
        return self.__ctx_def_map[name].file

    def get_yaml_api(self, name):
        """
        Get yaml api value from context definition map
        """
        return self.__ctx_def_map.get(name)

    def get_deduced_languages(self):
        """
        Returns the list of languages deduced from snippets info under "rules" section
        """
        return self.__deduced_languages

    def get_all_languages(self):
        """
        Returns the set of all languages: deduced languages merged with application supported languages
        """
        return set(self.__deduced_languages + default_config.languages)



# Copyright (c) 2022 PicsArt, Inc.
# All rights reserved. Use of this source code is governed by a
# MIT-style license that can be found in the LICENSE file.

import glob
import os

from setuptools import setup, find_packages

entry_point = (
    "cppbind = cppbind.runner:run_package"
)

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

with open("src/requirements.txt", "r") as f:
    requires = []
    for line in f:
        req = line.split("#", 1)[0].strip()
        if req and not req.startswith("--"):
            requires.append(req)

setup(
    name='cppbind',
    version="1.0.3",
    description="C++ bindings generator for various languages",
    author="Picsart",
    author_email="info@cppbind.io",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/PicsArt/cppbind",
    project_urls={
        "Documentation": "https://cppbind.io/",
    },

    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: C++",
        "Programming Language :: Kotlin",
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Code Generators",
    ],

    packages=find_packages('src/.', exclude=['test*']),
    package_dir={'': 'src'},
    # this is hack need to find proper way
    package_data={'cppbind': [os.path.relpath(f, 'src/cppbind/')
                              for f in glob.glob('src/cppbind/config/**/*', recursive=True)]},
    entry_points={"console_scripts": [entry_point]},
    install_requires=requires,
    python_requires='>=3.8',
)



# Copyright (c) 2022 PicsArt, Inc.
# All rights reserved. Use of this source code is governed by a
# MIT-style license that can be found in the LICENSE file.

"""
CppBind is a tool to generate c++ bindings for various languages.
"""
import datetime
import logging
import os
import sys
from logging import getLevelName

from cppbind.common.config import config as default_config, PROJECT_CONFIG_DIR
from cppbind.common.error import Error

BANNER_LOGO = """
  ______ .______   .______   .______    __  .__   __.  _______  
 /      ||   _  \  |   _  \  |   _  \  |  | |  \ |  | |       \ 
|  ,----'|  |_)  | |  |_)  | |  |_)  | |  | |   \|  | |  .--.  |
|  |     |   ___/  |   ___/  |   _  <  |  | |  . `  | |  |  |  |
|  `----.|  |      |  |      |  |_)  | |  | |  |\   | |  '--'  |
 \______|| _|      | _|      |______/  |__| |__| \__| |_______/ 
"""

LOGGING_FORMAT = '%(asctime)s:(%(levelname)s) %(message)s'
DATETIME_FORMAT = '%m/%d/%Y-%H:%M'
DATETIME_REGEX = r'^(0[1-9]|1[0-2])/([012][0-9]|3[0-1])/(20[0-9][0-9])-([01][0-9]|2[0-4]):([012345][0-9])'
LOG_LEVELS = ['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL']


def init_logger(log_level=None):
    log_level = log_level or default_config.logging.level
    log_kwargs = {}
    if default_config.logging.log_file:
        log_kwargs['filename'] = os.path.expanduser(default_config.logging.log_file)
    else:
        log_kwargs['stream'] = sys.stdout

    logging.basicConfig(**log_kwargs,
                        format=LOGGING_FORMAT,
                        level=getLevelName(log_level))


def find_prj_dir(dirname):
    if not os.path.isabs(dirname):
        if not os.path.isdir(dirname):
            dirname = os.path.abspath(os.path.join(PROJECT_CONFIG_DIR, dirname))
    return dirname


def current_datetime():
    """
    Returns formatted current date time in utc.
    Returns:
        str: Formatted result.
    """
    return datetime.date.strftime(datetime.datetime.utcnow(), DATETIME_FORMAT)



# Copyright (c) 2022 PicsArt, Inc.
# All rights reserved. Use of this source code is governed by a
# MIT-style license that can be found in the LICENSE file.

import copy
import types

import cppbind
import cppbind.converter
from cppbind.common.snippets_engine import (
    JINJA_UNIQUE_MARKER,
    SnippetsEngine,
)
from cppbind.common.type_info import create_type_info
from cppbind.cxx_exposed import (
    CXXClassExposedElement,
    CXXEnumExposedElement,
    CXXExposedType,
    CXXFunctionExposedElement,
    CXXMemberExposedElement
)
from cppbind.ir.ast import DIR_KIND_NAME, FILE_KIND_NAME
from cppbind.utils import get_public_attributes


SNIPPETS_ENGINE = None
GLOBAL_VARIABLES = {}
# variables below should be set after loading this module by calling set_language function
LANGUAGE = None


def set_language(language):
    global LANGUAGE
    LANGUAGE = language


def load_snippets_engine(runner, ctx_desc, platform, language):
    global SNIPPETS_ENGINE
    SNIPPETS_ENGINE = SnippetsEngine(runner, ctx_desc, platform, language)
    SNIPPETS_ENGINE.load()


def gen_init(ctx, ctx_desc, platform, language, *args, **kwargs):
    global SNIPPETS_ENGINE, GLOBAL_VARIABLES
    # load snippets

    load_snippets_engine(ctx.runner, ctx_desc, platform, language)

    GLOBAL_VARIABLES = SNIPPETS_ENGINE.do_actions({'vars': ctx.vars})


def make_def_context(ctx):
    def make():
        # helper variables
        banner_logo = cppbind.BANNER_LOGO
        new_line = cppbind.converter.NEW_LINE

        vars = ctx.vars

        return locals()

    context = make()
    context.update(GLOBAL_VARIABLES)
    return context


def make_dir_context(ctx):
    return make_def_context(ctx)


def make_file_context(ctx):
    return make_def_context(ctx)


def make_func_context(ctx):
    def make():
        overloading_index = ctx.node.overloading_index
        # capturing template related properties since we use single context with different template choice
        template_args_postfixes = ctx.template_args_postfixes

        cxx = CXXFunctionExposedElement(ctx.node.cxx_element, ctx.template_choice)

        return get_public_attributes(locals())

    context = make_def_context(ctx)
    context.update(make())
    return context


def make_method_context(ctx):
    def make():
        if ctx.setter:
            # setter is generated alongside with getter, setting template choice from getter context
            _setter_ctx = ctx.setter
            _setter_ctx.set_template_info(ctx.template_info)
            setter = make_method_context(_setter_ctx)

        return get_public_attributes(locals())

    context = make_func_context(ctx)
    context.update(make())
    return context


def make_constructor_context(ctx):
    return make_func_context(ctx)


def make_enum_context(ctx):
    def make():
        # helper variables
        cxx = CXXEnumExposedElement(ctx.node.cxx_element)

        return locals()

    context = make_def_context(ctx)
    context.update(make())
    return context


def make_class_context(ctx):
    def _make(ctx):
        def make():
            # for cases when type kind is invalid clang type does not give enough information
            # for such cases we use string type name
            _cxx_exposed_type = CXXExposedType(cxx_type=ctx.cxx_type_name,
                                               template_choice=ctx.template_choice,
                                               cxx_element=ctx.node.cxx_element)
            _type_info = create_type_info(ctx.runner, _cxx_exposed_type)

            cxx = CXXClassExposedElement(ctx.node.cxx_element, template_choice=ctx.template_choice)

            descendants = _type_info.descendants

            return get_public_attributes(locals())

        context = make_def_context(ctx)
        context.update(make())
        return context

    context = _make(ctx)
    return context


def make_member_context(ctx):
    def make():
        cxx = CXXMemberExposedElement(ctx.node.cxx_element, template_choice=ctx.template_choice)

        return get_public_attributes(locals())

    context = make_def_context(ctx)
    context.update(make())
    return context


def preprocess_scope(context, scope, info):
    context_scope = copy.copy(context)
    for sname in info.scopes:
        s = scope.create_scope(sname)
        context_scope[sname] = s
    if info.snippet_tmpl:
        scope.add(info.make_snippet(context_scope))
    if info.unique_snippet_tmpl:
        scope.add_unique(*str(info.unique_make_snippet(context_scope)).split(JINJA_UNIQUE_MARKER))


def preprocess_entry(context, builder, code_name):
    code_info = None
    if 'name' in context:
        code_info = SNIPPETS_ENGINE.get_code_info(f"{code_name}_{context['name']}")

    code_info = code_info or SNIPPETS_ENGINE.get_code_info(code_name)
    # continue processing if not an empty rule
    if code_info:
        for fs, info in code_info.items():
            fscope_name, scope_name = fs
            file_scope = get_file(context, builder, fscope_name)
            parent_scope = file_scope[scope_name]
            if parent_scope is not None:
                preprocess_scope(context, parent_scope, info)


def get_file(context, builder, fscope_name):
    file_info = SNIPPETS_ENGINE.get_file_info(fscope_name)
    file_name = file_info.get_file_name(context)

    return builder.get_file(file_name, init_func=lambda s: preprocess_scope(context, s, file_info))


def gen_entity(ctx, builder, api, node_kind_name):
    """A function for calling the right preprocess function"""

    # remove 'gen_'
    api_instruction_name = api[4:]

    if node_kind_name == DIR_KIND_NAME:
        context = make_dir_context(ctx)
    elif node_kind_name == FILE_KIND_NAME:
        context = make_file_context(ctx)
    elif node_kind_name in (
        'class',
        'class_template_partial_specialization',
        'class_template',
        'struct',
        'struct_template'
    ):
        context = make_class_context(ctx)
    elif node_kind_name == 'enum':
        context = make_enum_context(ctx)
    elif node_kind_name in (
        'function',
        'function_template'
    ):
        context = make_func_context(ctx)
    elif node_kind_name in (
        'method',
        'method_template'
    ):
        context = make_method_context(ctx)
    elif node_kind_name in (
        'constructor',
        'constructor_template',
    ):
        context = make_constructor_context(ctx)
    elif node_kind_name == 'field':
        context = make_member_context(ctx)
    else:
        context = make_def_context(ctx)

    preprocess_entry(context, builder, api_instruction_name)



# Copyright (c) 2022 PicsArt, Inc.
# All rights reserved. Use of this source code is governed by a
# MIT-style license that can be found in the LICENSE file.

"""
Module is intended to support modifications/augmentations to IR
"""

from cppbind.ir.ast import CXXNode, Node


class IRPostProcessor:
    """Class intended to do manipulations with IR"""

    def process_ir(self, ir):
        """Public method to do manipulations with ir"""
        self.__register_direct_descendants(ir)

        # return modified IR, since not always the changes can be done in-place
        return ir

    def __register_direct_descendants(self, ir):
        """Private method to register each node to its bases in IR"""

        for node in ir._get_all_nodes():
            if isinstance(node, CXXNode) and node.is_class_or_struct and node.api != Node.API_NONE:
                for base_node in node._base_type_specifier_nodes:
                    base_node.direct_descendants.add(node)
                    for template_node in base_node._template_spec_node_map.values():
                        # add also to specializations, as currently, we do not have a mechanism to identify whether
                        # it is template's descendant or specialization's
                        template_node.direct_descendants.add(node)



# Copyright (c) 2022 PicsArt, Inc.
# All rights reserved. Use of this source code is governed by a
# MIT-style license that can be found in the LICENSE file.

"""
"""
import itertools
import json
import types
from dataclasses import dataclass

from cached_property import cached_property

import cppbind.utils.clang as cutil
from cppbind import logging
from cppbind.ir import ElementKind
from cppbind.ir.ast import NodeType, Node
from cppbind.ir.utils import available_on
from cppbind.parser import TEMPLATE_TYPE_PARAMETER_KEY, TEMPLATE_NON_TYPE_PARAMETER_KEY
from cppbind.parser.cppbind_api_parser import (
    TEMPLATE_NAME_KEY,
    TEMPLATE_INSTANCE_KEY,
    TEMPLATE_INSTANCE_ARGS_KEY,
    TEMPLATE_KEY
)


@dataclass
class TemplateInfo:
    """Holds information about specialized template type's choice and template argument postfixes."""

    # template parameter - actual type mapping(e.g {'T': 'std::string', 'N': 'int'})
    choice: dict
    # template parameter argument postfixes retrieved from the api, will be none if not specified by user
    # these postfixes are used when constructing the name for template type in target languages, see object converter
    template_args_postfixes: list


class BaseContext:
    def __init__(self, runner, node=None):
        self.runner = runner
        self.node = node or runner.ir

    @cached_property
    def vars(self):
        return types.SimpleNamespace(**self.node.args)


class Context(BaseContext):

    def __init__(self, runner, node, template_info=None):
        super().__init__(runner, node)
        self.template_info = template_info

    @property
    def setter(self):
        """Find the context of the setter corresponding to the current getter context"""

        if self.node.api != 'gen_getter':
            return None

        search_api = 'gen_setter'
        name = self.node.spelling
        if name.lower().startswith('get'):
            name = name[3:].lstrip('_')
        search_names = {f"set_{name}", "set" + name[:1].upper() + name[1:],
                        'set' + ''.join(map(lambda n: n.capitalize(), name.split('_')))}

        node = self.node.find_adjacent_node(search_names, search_api)
        return self.runner.find_ctx_by_node(node)

    @property
    def getter(self):
        """Find the context of the getter corresponding to the current setter context"""

        if self.node.api != 'gen_setter':
            raise AttributeError(f"{self.__class__.__name__}.getter is invalid.")

        search_api = 'gen_getter'
        name = self.node.spelling
        if name.lower().startswith('set'):
            name = name[3:].lstrip('_')
        search_names = {f"get_{name}", "get" + name[:1].upper() + name[1:],
                        name, name[:1].lower() + name[1:]}

        node = self.node.find_adjacent_node(search_names, search_api)
        return self.runner.find_ctx_by_node(node)

    def set_template_info(self, template_info):
        self.template_info = template_info

    @property
    def template_choice(self):
        return self.template_info.choice if self.template_info else None

    @property
    def template_args_postfixes(self):
        return self.template_info.template_args_postfixes if self.template_info else None

    def lookup_ctx_by_name(self, name):
        return self.runner.get_context(name)

    @property
    @available_on(ElementKind.STRUCT_DECL,
                  ElementKind.CLASS_DECL,
                  ElementKind.CLASS_TEMPLATE,
                  ElementKind.CLASS_TEMPLATE_PARTIAL_SPECIALIZATION,
                  ElementKind.ENUM_DECL)
    def cxx_type_name(self):
        """
        Returns cxx type name. Main usage of this property is for template types. For templates clang's type information
        is not enough, so we retrieve some additional info from type's string representation
        Returns:
           str: Type name.
        """
        # in case of a template class - cursor type is TypeKind.INVALID,
        # that's why can't get type spelling from the
        # cursor using this approach instead
        # for example for the type a::Stack<T> full_displayname=a::Stack,
        # spelling=Stack, displayname=Stack<T>

        template_choice = self.template_choice or {}
        if self.node.cxx_element.is_templated:
            cxx_type_name = self.node.full_displayname
            return cutil.replace_template_choice(cxx_type_name, template_choice)
        return self.node.cxx_element.type.spelling


class RunRule:
    def __init__(self, ir, ctx_desc, platform, language):
        self.ir = ir
        self.ctx_desc = ctx_desc
        self.language = language
        self.platform = platform
        self.all_contexts = dict()
        # calling order should be such as that parent node processes first
        self.api_call_order = [
            # {'gen_class', 'gen_interface', 'gen_enum'},
            # {'gen_constructor'},
            # {'gen_method'},
            # {'gen_getter'},
            # {'gen_setter'},
            {}
        ]

    def run(self, rule, builder):  # noqa: C901

        # running order is defined by type of node
        # at first we run nodes for namespace/class/enums and then for methods

        # walk is traversing in dept first order
        self.allocate_all_contexts()

        # init all rules
        init_att_name = "gen_init"
        logging.debug(f"Initialising rule for {self.language} for {self.platform} platform.")
        func = getattr(rule, init_att_name)
        if func:
            func(BaseContext(self), self.ctx_desc, self.platform, self.language)

        # executes once for a type
        processed = dict()
        for calling_api in self.api_call_order:
            logging.debug(f"Calling APIs: {calling_api}")

            def _run_recursive(node, template_info=None):
                stack_added = False
                node_key = RunRule.__node_key(node, template_info)
                if node.api and (not calling_api or node.api in calling_api) and node_key not in processed:
                    ancestor = node.ancestor_with_api
                    ancestor_key = RunRule.__node_key(ancestor, template_info)
                    if ancestor_key in processed:
                        # for already called api resume builders scope stack
                        logging.debug(f"Restoring stack for {ancestor.displayname}.")
                        builder.restore_stacks(processed[ancestor_key])
                        logging.debug(
                            f"Restored stack {RunRule.__str_stacks(processed[ancestor_key])}."
                        )
                    # allocate scope
                    if node.type == NodeType.CLANG_NODE:
                        stack_added = True
                        builder.add_scope_stack()

                    # call api
                    self.call_api(rule, node, builder, template_info)
                    logging.debug(f"Capturing stack for {node.displayname}.")
                    processed[node_key] = builder.capture_stacks()
                    logging.debug(f"Captured stack {RunRule.__str_stacks(processed[node_key])}.")

                logging.debug(f"Processing children for {node.displayname}.")
                # iteration is defined by the IR
                for child in node:
                    if child.type == NodeType.CLANG_NODE:
                        # check if the node is template and generate code for each combination of template args
                        if child.kind in (ElementKind.CLASS_TEMPLATE,
                                          ElementKind.CLASS_TEMPLATE_PARTIAL_SPECIALIZATION,
                                          ElementKind.FUNCTION_TEMPLATE):
                            for _template_info in RunRule._get_template_infos(child, ignore_parents=True):
                                if template_info:
                                    # merge current node's context with parent
                                    # this might be revisited in #242
                                    # and will be changed after template nodes will be exposed during
                                    # ir post-processing
                                    _template_info.choice.update(template_info.choice)
                                _run_recursive(child, _template_info)
                        else:
                            _run_recursive(child, template_info)
                    else:
                        _run_recursive(child, template_info)
                logging.debug(f"End processing children for {node.displayname}.")

                if stack_added:
                    builder.pop_scope_stack()

            # create common scope for entire ir
            # we have to add this scope to be able to allow
            # to write into single file from multiple roots
            # since we do not have gen api for file
            # we register file and add one scope
            builder.add_scope_stack()

            _run_recursive(self.ir)

            builder.pop_scope_stack()

    def call_api(self, rule, node, builder, template_info=None):
        api = node.api
        if api == Node.API_NONE:
            return
        logging.debug(f"Call API: {api.lstrip(api)} on {node.displayname}")
        context = self.get_context(node.signature)
        # set current template context to generate code based on correct template choice
        context.set_template_info(template_info)
        rule.gen_entity(context, builder, api, node.kind_name)

    def get_context(self, type_name):
        return self.all_contexts.get(type_name)

    def find_ctx_by_node(self, node):
        """Find context by node"""
        if node:
            return self.get_context(node.signature)

    def allocate_all_contexts(self):
        logging.debug("Allocating context for all nodes")
        for node in self.ir.walk():
            if node.api not in (None, Node.API_NONE):
                ctx = Context(self, node)
                self.all_contexts.setdefault(node.signature, ctx)

                # this will be changed after template nodes will be exposed during ir post-processing
                if node.type == NodeType.CLANG_NODE and node.is_class_or_struct and node.cxx_element.is_templated:
                    # for template create a context also with only template type name(NOTE: this might not be used)
                    self.all_contexts.setdefault(cutil.template_type_name(node.signature), ctx)
                    # and for each specialization also create a context
                    for template_info in RunRule._get_template_infos(node):
                        context = Context(self, node, template_info)
                        self.all_contexts[context.cxx_type_name] = context

    @staticmethod
    def _get_template_combinations_from_instance(node, ignore_parents=False):
        """
        Returns list of TemplateInfos from the value of template_instance.
        Args:
            node(CXXNode): A template node.
            ignore_parents(bool): A flag to indicate whether parent node's choice should be added or not.
        Returns:
            List[TemplateInfo]: List of template info retrieved from template_instance variables.
        """
        node_template_arg = node.args.get(TEMPLATE_INSTANCE_KEY) or []
        parent = node.parent
        parameters = node.cxx_element.template_parameters
        choices = []
        parent_choices = []
        if not ignore_parents:
            if parent.type == NodeType.CLANG_NODE and parent.is_class_or_struct:
                # calling _get_template_combinations not _get_template_combinations_from_instance as user might
                # interchangeably use template and template_instance variables
                parent_choices = RunRule._get_template_infos(node.parent, ignore_parents)

        for spec in node_template_arg:
            arguments = cutil.get_template_arguments_from_str(f'<{spec.get(TEMPLATE_INSTANCE_ARGS_KEY)}>')
            i = 0
            choice = {}
            while True:
                parameter = parameters[i]
                if parameter.is_variadic:
                    # assuming variadic parameter is the last one
                    if i == len(parameters) - 1:
                        choice[parameter.spelling] = arguments[i:]
                        break
                    else:
                        raise RuntimeError(
                            "CPPBind supports variadic parameter only if it's the last one in the parameter list.")
                else:
                    choice[parameter.spelling] = arguments[i]
                if i == len(parameters) - 1:
                    break
                i += 1
            template_choice_name = [spec[TEMPLATE_NAME_KEY]] if TEMPLATE_NAME_KEY in spec else None
            if parent_choices:
                choices += [TemplateInfo(choice={**parent_choice.choice, **choice},
                                         template_args_postfixes=template_choice_name) for parent_choice
                            in parent_choices]
            else:
                choices.append(TemplateInfo(choice=choice, template_args_postfixes=template_choice_name))

        return choices

    @staticmethod
    def _get_template_infos(node, ignore_parents=False):
        """
        Returns list of TemplateInfos from the value of template or template_instance.
        Args:
            node(CXXNode): A template node.
            ignore_parents(bool): A flag to indicate whether parent node's choice should be added or not.
        Returns:
            List[TemplateInfo]: List of template info retrieved from template or template_instance variables.
        """
        # this function logic will be moved when templates will be exposed during post-processing
        if TEMPLATE_INSTANCE_KEY in node.args:
            return RunRule._get_template_combinations_from_instance(node, ignore_parents)

        parent = node.parent
        # for template choice we need to consider all template parent types
        template_arg = {}

        # current node template argument
        node_template_arg = node.args.get(TEMPLATE_KEY) or {}
        template_arg.update(node_template_arg)

        parent_choices = {}
        if not ignore_parents:
            if parent.type == NodeType.CLANG_NODE and parent.is_class_or_struct:
                # calling _get_template_combinations as user might
                # interchangeably use template and template_instance variables
                parent_choices = RunRule._get_template_infos(node.parent, ignore_parents)

        # reorder user provided template  to match the template type/function order
        node_template_arg = {param.spelling: node_template_arg[param.spelling] for param in
                             node.cxx_element.template_parameters}

        all_possible_args = list(itertools.product(*node_template_arg.values()))
        template_keys = node_template_arg.keys()
        all_contexts = []
        for _, combination in enumerate(all_possible_args):
            choice = []
            choice_names = []
            for template_arg_info in combination:
                choice.append(template_arg_info.get(TEMPLATE_TYPE_PARAMETER_KEY, template_arg_info.get(
                    TEMPLATE_NON_TYPE_PARAMETER_KEY)))
                choice_names.append(template_arg_info.get(TEMPLATE_NAME_KEY))
            _template_choice = dict(zip(template_keys, choice))
            if parent_choices:
                all_contexts += [
                    TemplateInfo(choice={**_template_choice, **parent_choice.choice},
                                 template_args_postfixes=choice_names) for
                    parent_choice in parent_choices]
            else:
                all_contexts.append(TemplateInfo(choice=_template_choice, template_args_postfixes=choice_names))
        return all_contexts

    @staticmethod
    def __node_key(node, template_info):
        node_key = (node,)
        if node and node.type == NodeType.CLANG_NODE and node.cxx_element.is_templated:
            node_key = (node, json.dumps(template_info.__dict__))
        return node_key

    @staticmethod
    def __str_stacks(capture_data):
        # first lang stack data
        stack_data = capture_data
        return f"{[s.keys() for fs in stack_data.values() for s in fs]}"
